use crate::auth::AuthSession;
use crate::database::GetConnection;
use crate::models::user::{CreateUser, User, UserCredentials};
use crate::result::{ResponseResult, Success};
use crate::services::user_service::UserService;
use crate::{bad_request, internal_server_error, ok};
use axum::extract::State;
use axum::Json;
use axum_valid::Valid;
// router!(
//     "/auth",
//     routes!(
//         post "/login" => login::<Pool>,
//         post "/register" => register::<Pool>
//     ),
//     Pool: Clone, Send, Sync, GetConnection -> UserService
// );

pub fn router<Pool: Clone + Send + Sync + GetConnection + 'static>(
) -> axum::Router<UserService<Pool>> {
    axum::Router::new().nest(
        "/auth",
        axum::Router::new()
            .route("/login", axum::routing::post(login::<Pool>))
            .route("/register", axum::routing::post(register::<Pool>)),
    )
}

async fn login<Pool>(
    mut auth_session: AuthSession<Pool>,
    Valid(Json(credentials)): Valid<Json<UserCredentials>>,
) -> ResponseResult
where
    Pool: GetConnection,
{
    let user = match auth_session.authenticate(credentials).await {
        Ok(Some(user)) => user,
        Ok(None) => return bad_request!("Invalid credentials"),
        Err(error) => return internal_server_error!(error),
    };
    auth_session.login(&user).await?;
    ok!()
}

async fn register<Pool>(
    State(service): State<UserService<Pool>>,
    Valid(Json(create_user)): Valid<Json<CreateUser>>,
) -> ResponseResult<User>
where
    Pool: GetConnection,
{
    service
        .insert(create_user)
        .await
        .map(Success::Created)
        .map_err(Into::into)
}

#[cfg(test)]
mod tests {
    use crate::create_app;
    use crate::error::ErrorResponseBody;
    use crate::models::user::{CreateUser, User, UserRole};
    use crate::test::{create_test_containers_pool, create_test_pool, BuildJson, DeserializeInto};
    use axum::http::request::Builder;
    use axum::http::{Request, StatusCode};
    use secrecy::ExposeSecret;
    use serde::ser::SerializeStruct;
    use serde::{Serialize, Serializer};
    use tower::ServiceExt;

    fn register() -> Builder {
        Request::builder().uri("/auth/register").method("POST")
    }

    #[tokio::test]
    async fn test_register_created() {
        let pool = create_test_pool().await.unwrap();
        let app = create_app(pool);
        let create_user = CreateUser::new("test@email.com", "password");
        let create_email = create_user.email.clone();

        let response = app
            .oneshot(register().json(create_user).unwrap())
            .await
            .unwrap();

        let status = response.status();
        let user: User = response.deserialize_into().await.unwrap();

        assert_eq!(status, StatusCode::CREATED);
        assert_eq!(user.email, create_email);
        assert_eq!(user.role, UserRole::default() as i16);
        assert!(!user.hash.is_empty());
        assert!(!user.salt.is_empty());
    }

    #[tokio::test]
    async fn test_register_email_already_registered() {
        let test_container = create_test_containers_pool().await.unwrap();
        let app = create_app(test_container.pool);

        let create_user = CreateUser::new("test@email.com", "password");

        let call = || async {
            app.clone()
                .oneshot(register().json(create_user.clone()).unwrap())
                .await
                .unwrap()
        };
        let created = call().await;
        let conflicted = call().await;

        let conflicted_status = conflicted.status();

        let response: ErrorResponseBody = conflicted.deserialize_into().await.unwrap();

        assert_eq!(created.status(), StatusCode::CREATED);
        assert_eq!(conflicted_status, StatusCode::CONFLICT);
        assert!(
            response.error.contains("user_pkey"),
            "Error: {}",
            response.error
        );
    }

    impl Serialize for CreateUser {
        fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
        where
            S: Serializer,
        {
            let mut s = serializer.serialize_struct("CreateUser", 2)?;
            s.serialize_field("email", &self.email)?;
            s.serialize_field("password", &self.password.expose_secret())?;
            s.serialize_field("role", &self.role)?;
            s.end()
        }
    }
}
